#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "AkelEdit.h"
#include "AkelDLL.h"
#include "Resources\Resource.h"


//Defines
#define DLLA_AUTOSCROLL_HSCROLL      1
#define DLLA_AUTOSCROLL_VSCROLL      2
#define DLLA_AUTOSCROLL_LINEHSCROLL  3
#define DLLA_AUTOSCROLL_LINEVSCROLL  4
#define DLLA_AUTOSCROLL_GOAUTOSCROLL 5
#define DLLA_AUTOSCROLL_GOSYNC       6

#define STRID_AUTOSTEP          1
#define STRID_ENABLE            2
#define STRID_STEPTIME          3
#define STRID_STEPWIDTH         4
#define STRID_HSYNCHRONIZATION  5
#define STRID_VSYNCHRONIZATION  6
#define STRID_FRAMEMDI          7
#define STRID_SPLITPANE         8
#define STRID_NOSCROLL          9
#define STRID_UNDO              10
#define STRID_REDO              11
#define STRID_SELECTALL         12
#define STRID_OK                13
#define STRID_CANCEL            14

#define NS_UNDO      0x1
#define NS_REDO      0x2
#define NS_SELECTALL 0x4

//Synchronization
#define SNC_MDI         0x00000001  //Synchronize scroll of MDI windows.
#define SNC_CLONE       0x00000002  //Synchronize scroll of split windows.

#ifndef EM_GETSCROLLPOS
  #define EM_GETSCROLLPOS (WM_USER + 221)
#endif
#ifndef EM_SETSCROLLPOS
  #define EM_SETSCROLLPOS (WM_USER + 222)
#endif

//Functions prototypes
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void SetScrollTimer();
void KillScrollTimer();
VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime);
void ReadOptionsA();
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA();
void SaveOptionsW();
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();
WNDPROCDATA *NewMainProcData=NULL;
WNDPROCDATA *NewFrameProcData=NULL;

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
HWND hMdiClient;
HICON hMainIcon;
BOOL bOldWindows;
BOOL bMDI;
BOOL bAkelEdit;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
BOOL bDoInit=TRUE;
HWND hWndScrolling;
DWORD dwScrollTimer=0;
BOOL bStepEnable=TRUE;
DWORD dwStepTime=50;
int nStepWidth=1;
DWORD dwHSync=0;
DWORD dwVSync=0;
DWORD dwNoScroll=0;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="AutoScroll";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (pd->lParam)
  {
    if (IsExtCallParamValid(pd->lParam, 1))
    {
      if (pd->bAkelEdit)
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_HSCROLL)
        {
          int nValue=0;

          if (IsExtCallParamValid(pd->lParam, 2))
            nValue=GetExtCallParam(pd->lParam, 2);

          if (nValue)
          {
            SendMessage(pd->hWndEdit, AEM_SCROLL, SB_HORZ, nValue);
          }
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_VSCROLL)
        {
          int nValue=0;

          if (IsExtCallParamValid(pd->lParam, 2))
            nValue=GetExtCallParam(pd->lParam, 2);

          if (nValue)
          {
            SendMessage(pd->hWndEdit, AEM_SCROLL, SB_VERT, nValue);
          }
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_LINEHSCROLL)
        {
          AESELECTION aes;
          AECHARINDEX ciCaret;
          AESCROLLTOPOINT stp;
          DWORD dwScrollResult;
          int nValue1=0;
          int nValue2=0;

          if (IsExtCallParamValid(pd->lParam, 2))
            nValue1=GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            nValue2=GetExtCallParam(pd->lParam, 3);

          if (nValue1)
          {
            if (nValue2 == 0)
            {
              SendMessage(pd->hWndEdit, AEM_LINESCROLL, AESB_HORZ, nValue1);
            }
            else if (nValue2 == 1)
            {
              //Test scroll to caret
              stp.dwFlags=AESC_TEST|AESC_POINTCARET|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
              stp.nOffsetX=0;
              stp.nOffsetY=0;
              dwScrollResult=SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);

              if (SendMessage(pd->hWndEdit, AEM_LINESCROLL, AESB_HORZ, nValue1))
              {
                if (!(dwScrollResult & AECSE_SCROLLEDX) && !(dwScrollResult & AECSE_SCROLLEDY))
                {
                  SendMessage(pd->hWndEdit, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);
                  ciCaret.nCharInLine=max(ciCaret.nCharInLine + nValue1, 0);

                  //Set new caret position
                  aes.crSel.ciMin=ciCaret;
                  aes.crSel.ciMax=ciCaret;
                  aes.dwFlags=AESELT_LOCKSCROLL;
                  SendMessage(pd->hWndEdit, AEM_SETSEL, (WPARAM)NULL, (LPARAM)&aes);
                }
              }
            }
            else if (nValue2 == 2)
            {
              SendMessage(pd->hWndEdit, AEM_LINESCROLL, AESB_HORZ, nValue1);

              //Test scroll to caret
              stp.dwFlags=AESC_TEST|AESC_POINTCARET|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
              stp.nOffsetX=0;
              stp.nOffsetY=0;
              dwScrollResult=SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);

              if (dwScrollResult & AECSE_SCROLLEDX)
              {
                SendMessage(pd->hWndEdit, AEM_GETINDEX, AEGI_CARETCHAR, (LPARAM)&ciCaret);
                ciCaret.nCharInLine=max(ciCaret.nCharInLine + nValue1, 0);

                //Set new caret position
                aes.crSel.ciMin=ciCaret;
                aes.crSel.ciMax=ciCaret;
                aes.dwFlags=AESELT_LOCKSCROLL;
                SendMessage(pd->hWndEdit, AEM_SETSEL, (WPARAM)NULL, (LPARAM)&aes);

                //Scroll to caret
                stp.dwFlags=AESC_POINTCARET|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
                stp.nOffsetX=0;
                stp.nOffsetY=0;
                SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);
              }
              else if (dwScrollResult & AECSE_SCROLLEDY)
              {
                //Scroll to caret
                stp.dwFlags=AESC_POINTCARET|AESC_OFFSETRECTDIVY;
                stp.nOffsetX=0;
                stp.nOffsetY=2;
                SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);
              }
            }
          }
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_LINEVSCROLL)
        {
          AESELECTION aes;
          AECHARINDEX ciCaret;
          AESCROLLTOPOINT stp;
          POINT pt;
          DWORD dwAlign=0;
          DWORD dwScrollResult;
          int nValue1=0;
          int nValue2=0;

          if (IsExtCallParamValid(pd->lParam, 2))
            nValue1=GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            nValue2=GetExtCallParam(pd->lParam, 3);

          if (nValue1)
          {
            if (nValue1 > 0)
              dwAlign=AESB_ALIGNTOP;
            else if (nValue1 < 0)
              dwAlign=AESB_ALIGNBOTTOM;

            if (nValue2 == 0)
            {
              SendMessage(pd->hWndEdit, AEM_LINESCROLL, AESB_VERT|dwAlign, nValue1);
            }
            else if (nValue2 == 1)
            {
              //Test scroll to caret
              stp.dwFlags=AESC_TEST|AESC_POINTCARET|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
              stp.nOffsetX=0;
              stp.nOffsetY=0;
              dwScrollResult=SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);

              if (SendMessage(pd->hWndEdit, AEM_LINESCROLL, AESB_VERT|dwAlign, nValue1))
              {
                if (!(dwScrollResult & AECSE_SCROLLEDY))
                {
                  SendMessage(pd->hWndEdit, AEM_GETCARETPOS, 0, (WPARAM)&pt);
                  pt.x=SendMessage(pd->hWndEdit, AEM_GETCARETHORZINDENT, 0, 0);
                  pt.y+=nValue1 * SendMessage(pd->hWndEdit, AEM_GETCHARSIZE, AECS_HEIGHT, 0);
                  SendMessage(pd->hWndEdit, AEM_CONVERTPOINT, AECPT_GLOBALTOCLIENT, (LPARAM)&pt);
                  SendMessage(pd->hWndEdit, AEM_CHARFROMPOS, (WPARAM)&pt, (LPARAM)&ciCaret);

                  //Set new caret position
                  aes.crSel.ciMin=ciCaret;
                  aes.crSel.ciMax=ciCaret;
                  aes.dwFlags=AESELT_LOCKSCROLL|AESELT_NOCARETHORZINDENT;
                  SendMessage(pd->hWndEdit, AEM_SETSEL, (WPARAM)NULL, (LPARAM)&aes);
                }
              }
            }
            else if (nValue2 == 2)
            {
              SendMessage(pd->hWndEdit, AEM_LINESCROLL, AESB_VERT|dwAlign, nValue1);

              //Test scroll to caret
              stp.dwFlags=AESC_TEST|AESC_POINTCARET|AESC_OFFSETCHARX|AESC_OFFSETCHARY;
              stp.nOffsetX=0;
              stp.nOffsetY=0;
              dwScrollResult=SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);

              if (dwScrollResult & AECSE_SCROLLEDY)
              {
                SendMessage(pd->hWndEdit, AEM_GETCARETPOS, 0, (WPARAM)&pt);
                pt.x=SendMessage(pd->hWndEdit, AEM_GETCARETHORZINDENT, 0, 0);
                pt.y+=nValue1 * SendMessage(pd->hWndEdit, AEM_GETCHARSIZE, AECS_HEIGHT, 0);
                SendMessage(pd->hWndEdit, AEM_CONVERTPOINT, AECPT_GLOBALTOCLIENT, (LPARAM)&pt);
                SendMessage(pd->hWndEdit, AEM_CHARFROMPOS, (WPARAM)&pt, (LPARAM)&ciCaret);

                //Set new caret position
                aes.crSel.ciMin=ciCaret;
                aes.crSel.ciMax=ciCaret;
                aes.dwFlags=AESELT_NOCARETHORZINDENT;
                SendMessage(pd->hWndEdit, AEM_SETSEL, (WPARAM)NULL, (LPARAM)&aes);
              }
              else if (dwScrollResult & AECSE_SCROLLEDX)
              {
                //Scroll to caret
                stp.dwFlags=AESC_POINTCARET|AESC_OFFSETRECTDIVX;
                stp.nOffsetX=3;
                stp.nOffsetY=0;
                SendMessage(pd->hWndEdit, AEM_SCROLLTOPOINT, 0, (LPARAM)&stp);
              }
            }
          }
        }
      }
      bDoInit=FALSE;

      if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_GOAUTOSCROLL ||
          GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_GOSYNC)
      {
        int nValue1=0;
        int nValue2=0;

        if (IsExtCallParamValid(pd->lParam, 2))
          nValue1=GetExtCallParam(pd->lParam, 2);
        if (IsExtCallParamValid(pd->lParam, 3))
          nValue2=GetExtCallParam(pd->lParam, 3);

        if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_GOAUTOSCROLL)
        {
          if (nValue1 && nValue2)
          {
            bStepEnable=TRUE;
            dwStepTime=nValue1;
            nStepWidth=nValue2;
          }
          else bStepEnable=FALSE;
        }
        else if (GetExtCallParam(pd->lParam, 1) == DLLA_AUTOSCROLL_GOSYNC)
        {
          dwHSync=nValue1;
          dwVSync=nValue2;
        }

        if (bInitMain)
        {
          UninitMain();
          hWndScrolling=NULL;
        }
        bDoInit=TRUE;
      }
    }

    //Stay in memory, and don't change active status
    pd->nUnload=UD_NONUNLOAD_UNCHANGE;
  }

  if (bDoInit)
  {
    if (bInitMain)
    {
      UninitMain();
      hWndScrolling=NULL;

      //Unload plugin
      pd->nUnload=UD_UNLOAD;
    }
    else
    {
      hWndScrolling=pd->hWndEdit;
      InitMain();

      //Stay in memory, and show as active
      pd->nUnload=UD_NONUNLOAD_ACTIVE;
    }
  }
}

extern "C" void __declspec(dllexport) Settings(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=FALSE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);

  if (bOldWindows)
    DialogBoxA(hInstanceDLL, MAKEINTRESOURCEA(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcA);
  else
    DialogBoxW(hInstanceDLL, MAKEINTRESOURCEW(IDD_SETUP), hMainWnd, (DLGPROC)SetupDlgProcW);

  //If plugin already loaded, stay in memory, but show as non-active
  if (pd->bInMemory) pd->nUnload=UD_NONUNLOAD_NONACTIVE;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (uMsg == AKDN_EDIT_ONFINISH)
  {
    if (hWndScrolling == (HWND)wParam)
    {
      KillScrollTimer();
      hWndScrolling=NULL;
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDM_EDIT_SELECTALL && (dwNoScroll & NS_SELECTALL))
    {
      EDITINFO ei;
      LRESULT lResult;

      if (bAkelEdit)
      {
        if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
        {
          SendMessage(ei.hWndEdit, AEM_LOCKSCROLL, SB_BOTH, TRUE);
          lResult=NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
          SendMessage(ei.hWndEdit, AEM_LOCKSCROLL, SB_BOTH, FALSE);
          return lResult;
        }
      }
    }
  }
  if (lResult=EditParentMessages(hWnd, uMsg, wParam, lParam))
    return lResult;

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  LRESULT lResult;

  if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if ((((NMHDR *)lParam)->code == AEN_HSCROLL && (dwHSync & SNC_MDI)) ||
          (((NMHDR *)lParam)->code == AEN_VSCROLL && (dwVSync & SNC_MDI)))
      {
        AENSCROLL *aens=(AENSCROLL *)lParam;
        HWND hWndChildFrame;
        HWND hWndChildEdit;
        POINT pt;
        static BOOL bLock=FALSE;

        if (!bLock)
        {
          bLock=TRUE;
          hWndChildFrame=GetWindow(hMdiClient, GW_CHILD);

          while (hWndChildFrame)
          {
            if (hWndChildFrame != hWnd)
            {
              hWndChildEdit=GetDlgItem(hWndChildFrame, ID_EDIT);

              SendMessage(hWndChildEdit, AEM_GETSCROLLPOS, 0, (LPARAM)&pt);
              if (((NMHDR *)lParam)->code == AEN_HSCROLL)
                pt.x+=aens->nPosNew - aens->nPosOld;
              else
                pt.y+=aens->nPosNew - aens->nPosOld;
              SendMessage(hWndChildEdit, AEM_SETSCROLLPOS, 0, (LPARAM)&pt);
            }
            hWndChildFrame=GetNextWindow(hWndChildFrame, GW_HWNDNEXT);
          }
          bLock=FALSE;
        }
      }
    }
  }
  if (lResult=EditParentMessages(hWnd, uMsg, wParam, lParam))
    return lResult;

  //Call next procedure
  return NewFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

LRESULT CALLBACK EditParentMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_NOTIFY)
  {
    if (wParam == ID_EDIT)
    {
      if ((((NMHDR *)lParam)->code == AEN_HSCROLL && (dwHSync & SNC_CLONE)) ||
          (((NMHDR *)lParam)->code == AEN_VSCROLL && (dwVSync & SNC_CLONE)))
      {
        AENSCROLL *aens=(AENSCROLL *)lParam;
        HWND hWndMaster;
        HWND hWndChildEdit;
        POINT pt;
        int nClone=0;
        static BOOL bLock=FALSE;

        if (!bLock)
        {
          bLock=TRUE;

          if (hWndMaster=(HWND)SendMessage(aens->hdr.hwndFrom, AEM_GETMASTER, 0, 0))
          {
            hWndChildEdit=hWndMaster;

            while (hWndChildEdit)
            {
              if (hWndChildEdit != aens->hdr.hwndFrom)
              {
                SendMessage(hWndChildEdit, AEM_GETSCROLLPOS, 0, (LPARAM)&pt);
                if (((NMHDR *)lParam)->code == AEN_HSCROLL)
                  pt.x+=aens->nPosNew - aens->nPosOld;
                else
                  pt.y+=aens->nPosNew - aens->nPosOld;
                SendMessage(hWndChildEdit, AEM_SETSCROLLPOS, 0, (LPARAM)&pt);
              }
              hWndChildEdit=(HWND)SendMessage(hWndMaster, AEM_GETCLONE, nClone++, 0);
            }
          }
          bLock=FALSE;
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGING)
      {
        AENTEXTCHANGE *aentc=(AENTEXTCHANGE *)lParam;

        if (((aentc->dwType & AETCT_UNDO) && (dwNoScroll & NS_UNDO)) ||
            ((aentc->dwType & AETCT_REDO) && (dwNoScroll & NS_REDO)))
        {
          SendMessage(aentc->hdr.hwndFrom, AEM_LOCKSCROLL, SB_BOTH, TRUE);
        }
      }
      else if (((NMHDR *)lParam)->code == AEN_TEXTCHANGED)
      {
        AENTEXTCHANGE *aentc=(AENTEXTCHANGE *)lParam;

        if (((aentc->dwType & AETCT_UNDO) && (dwNoScroll & NS_UNDO)) ||
            ((aentc->dwType & AETCT_REDO) && (dwNoScroll & NS_REDO)))
        {
          SendMessage(aentc->hdr.hwndFrom, AEM_LOCKSCROLL, SB_BOTH, FALSE);

          if (!(aentc->dwType & AETCT_DELETEALL))
          {
            SendMessage(aentc->hdr.hwndFrom, EM_SCROLLCARET, 0, 0);
          }
        }
      }
    }
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndStepEnable;
  static HWND hWndStepTime;
  static HWND hWndStepWidth;
  static HWND hWndHSyncMDI;
  static HWND hWndHSyncClone;
  static HWND hWndVSyncMDI;
  static HWND hWndVSyncClone;
  static HWND hWndUndo;
  static HWND hWndRedo;
  static HWND hWndSelectAll;
  BOOL bState;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndStepEnable=GetDlgItem(hDlg, IDC_STEPENABLE);
    hWndStepTime=GetDlgItem(hDlg, IDC_STEPTIME);
    hWndStepWidth=GetDlgItem(hDlg, IDC_STEPWIDTH);
    hWndHSyncMDI=GetDlgItem(hDlg, IDC_HSYNC_MDI);
    hWndHSyncClone=GetDlgItem(hDlg, IDC_HSYNC_CLONE);
    hWndVSyncMDI=GetDlgItem(hDlg, IDC_VSYNC_MDI);
    hWndVSyncClone=GetDlgItem(hDlg, IDC_VSYNC_CLONE);
    hWndUndo=GetDlgItem(hDlg, IDC_NOSCROLL_UNDO);
    hWndRedo=GetDlgItem(hDlg, IDC_NOSCROLL_REDO);
    hWndSelectAll=GetDlgItem(hDlg, IDC_NOSCROLL_SELECTALL);

    SetDlgItemTextA(hDlg, IDC_AUTOSTEP_GROUP, GetLangStringA(wLangSystem, STRID_AUTOSTEP));
    SetDlgItemTextA(hDlg, IDC_STEPENABLE, GetLangStringA(wLangSystem, STRID_ENABLE));
    SetDlgItemTextA(hDlg, IDC_STEPTIME_LABEL, GetLangStringA(wLangSystem, STRID_STEPTIME));
    SetDlgItemTextA(hDlg, IDC_STEPWIDTH_LABEL, GetLangStringA(wLangSystem, STRID_STEPWIDTH));
    SetDlgItemTextA(hDlg, IDC_HSYNC_GROUP, GetLangStringA(wLangSystem, STRID_HSYNCHRONIZATION));
    SetDlgItemTextA(hDlg, IDC_HSYNC_MDI, GetLangStringA(wLangSystem, STRID_FRAMEMDI));
    SetDlgItemTextA(hDlg, IDC_HSYNC_CLONE, GetLangStringA(wLangSystem, STRID_SPLITPANE));
    SetDlgItemTextA(hDlg, IDC_VSYNC_GROUP, GetLangStringA(wLangSystem, STRID_VSYNCHRONIZATION));
    SetDlgItemTextA(hDlg, IDC_VSYNC_MDI, GetLangStringA(wLangSystem, STRID_FRAMEMDI));
    SetDlgItemTextA(hDlg, IDC_VSYNC_CLONE, GetLangStringA(wLangSystem, STRID_SPLITPANE));
    SetDlgItemTextA(hDlg, IDC_NOSCROLL_GROUP, GetLangStringA(wLangSystem, STRID_NOSCROLL));
    SetDlgItemTextA(hDlg, IDC_NOSCROLL_UNDO, GetLangStringA(wLangSystem, STRID_UNDO));
    SetDlgItemTextA(hDlg, IDC_NOSCROLL_REDO, GetLangStringA(wLangSystem, STRID_REDO));
    SetDlgItemTextA(hDlg, IDC_NOSCROLL_SELECTALL, GetLangStringA(wLangSystem, STRID_SELECTALL));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));
    SetDlgItemTextA(hDlg, IDCANCEL, GetLangStringA(wLangSystem, STRID_CANCEL));

    SendMessage(hWndStepEnable, BM_SETCHECK, bStepEnable, 0);
    SendMessage(hWndStepTime, EM_LIMITTEXT, 5, 0);
    SendMessage(hWndStepWidth, EM_LIMITTEXT, 5, 0);
    SetDlgItemInt(hDlg, IDC_STEPTIME, dwStepTime, FALSE);
    SetDlgItemInt(hDlg, IDC_STEPWIDTH, nStepWidth, TRUE);

    if (dwHSync & SNC_MDI)
      SendMessage(hWndHSyncMDI, BM_SETCHECK, BST_CHECKED, 0);
    if (dwHSync & SNC_CLONE)
      SendMessage(hWndHSyncClone, BM_SETCHECK, BST_CHECKED, 0);
    if (dwVSync & SNC_MDI)
      SendMessage(hWndVSyncMDI, BM_SETCHECK, BST_CHECKED, 0);
    if (dwVSync & SNC_CLONE)
      SendMessage(hWndVSyncClone, BM_SETCHECK, BST_CHECKED, 0);

    if (dwNoScroll & NS_UNDO)
      SendMessage(hWndUndo, BM_SETCHECK, BST_CHECKED, 0);
    if (dwNoScroll & NS_REDO)
      SendMessage(hWndRedo, BM_SETCHECK, BST_CHECKED, 0);
    if (dwNoScroll & NS_SELECTALL)
      SendMessage(hWndSelectAll, BM_SETCHECK, BST_CHECKED, 0);

    if (!bAkelEdit || !bMDI)
    {
      EnableWindow(hWndHSyncMDI, FALSE);
      EnableWindow(hWndVSyncMDI, FALSE);
    }
    if (!bAkelEdit)
    {
      EnableWindow(hWndHSyncClone, FALSE);
      EnableWindow(hWndVSyncClone, FALSE);

      EnableWindow(hWndUndo, FALSE);
      EnableWindow(hWndRedo, FALSE);
      EnableWindow(hWndSelectAll, FALSE);
    }
    SendMessage(hDlg, WM_COMMAND, IDC_STEPENABLE, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_STEPENABLE)
    {
      bState=SendMessage(hWndStepEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndStepTime, bState);
      EnableWindow(hWndStepWidth, bState);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      bStepEnable=SendMessage(hWndStepEnable, BM_GETCHECK, 0, 0);
      dwStepTime=GetDlgItemInt(hDlg, IDC_STEPTIME, NULL, FALSE);
      nStepWidth=GetDlgItemInt(hDlg, IDC_STEPWIDTH, NULL, TRUE);

      dwHSync=0;
      if (SendMessage(hWndHSyncMDI, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwHSync|=SNC_MDI;
      if (SendMessage(hWndHSyncClone, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwHSync|=SNC_CLONE;

      dwVSync=0;
      if (SendMessage(hWndVSyncMDI, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwVSync|=SNC_MDI;
      if (SendMessage(hWndVSyncClone, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwVSync|=SNC_CLONE;

      dwNoScroll=0;
      if (SendMessage(hWndUndo, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwNoScroll|=NS_UNDO;
      if (SendMessage(hWndRedo, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwNoScroll|=NS_REDO;
      if (SendMessage(hWndSelectAll, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwNoScroll|=NS_SELECTALL;

      SaveOptionsA();
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        if (!hWndScrolling)
        {
          EDITINFO ei;

          if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
            hWndScrolling=ei.hWndEdit;
        }
        UninitMain();
        InitMain();
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}

BOOL CALLBACK SetupDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndStepEnable;
  static HWND hWndStepTime;
  static HWND hWndStepWidth;
  static HWND hWndHSyncMDI;
  static HWND hWndHSyncClone;
  static HWND hWndVSyncMDI;
  static HWND hWndVSyncClone;
  static HWND hWndUndo;
  static HWND hWndRedo;
  static HWND hWndSelectAll;
  BOOL bState;

  if (uMsg == WM_INITDIALOG)
  {
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)hMainIcon);
    hWndStepEnable=GetDlgItem(hDlg, IDC_STEPENABLE);
    hWndStepTime=GetDlgItem(hDlg, IDC_STEPTIME);
    hWndStepWidth=GetDlgItem(hDlg, IDC_STEPWIDTH);
    hWndHSyncMDI=GetDlgItem(hDlg, IDC_HSYNC_MDI);
    hWndHSyncClone=GetDlgItem(hDlg, IDC_HSYNC_CLONE);
    hWndVSyncMDI=GetDlgItem(hDlg, IDC_VSYNC_MDI);
    hWndVSyncClone=GetDlgItem(hDlg, IDC_VSYNC_CLONE);
    hWndUndo=GetDlgItem(hDlg, IDC_NOSCROLL_UNDO);
    hWndRedo=GetDlgItem(hDlg, IDC_NOSCROLL_REDO);
    hWndSelectAll=GetDlgItem(hDlg, IDC_NOSCROLL_SELECTALL);

    SetDlgItemTextW(hDlg, IDC_AUTOSTEP_GROUP, GetLangStringW(wLangSystem, STRID_AUTOSTEP));
    SetDlgItemTextW(hDlg, IDC_STEPENABLE, GetLangStringW(wLangSystem, STRID_ENABLE));
    SetDlgItemTextW(hDlg, IDC_STEPTIME_LABEL, GetLangStringW(wLangSystem, STRID_STEPTIME));
    SetDlgItemTextW(hDlg, IDC_STEPWIDTH_LABEL, GetLangStringW(wLangSystem, STRID_STEPWIDTH));
    SetDlgItemTextW(hDlg, IDC_HSYNC_GROUP, GetLangStringW(wLangSystem, STRID_HSYNCHRONIZATION));
    SetDlgItemTextW(hDlg, IDC_HSYNC_MDI, GetLangStringW(wLangSystem, STRID_FRAMEMDI));
    SetDlgItemTextW(hDlg, IDC_HSYNC_CLONE, GetLangStringW(wLangSystem, STRID_SPLITPANE));
    SetDlgItemTextW(hDlg, IDC_VSYNC_GROUP, GetLangStringW(wLangSystem, STRID_VSYNCHRONIZATION));
    SetDlgItemTextW(hDlg, IDC_VSYNC_MDI, GetLangStringW(wLangSystem, STRID_FRAMEMDI));
    SetDlgItemTextW(hDlg, IDC_VSYNC_CLONE, GetLangStringW(wLangSystem, STRID_SPLITPANE));
    SetDlgItemTextW(hDlg, IDC_NOSCROLL_GROUP, GetLangStringW(wLangSystem, STRID_NOSCROLL));
    SetDlgItemTextW(hDlg, IDC_NOSCROLL_UNDO, GetLangStringW(wLangSystem, STRID_UNDO));
    SetDlgItemTextW(hDlg, IDC_NOSCROLL_REDO, GetLangStringW(wLangSystem, STRID_REDO));
    SetDlgItemTextW(hDlg, IDC_NOSCROLL_SELECTALL, GetLangStringW(wLangSystem, STRID_SELECTALL));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));
    SetDlgItemTextW(hDlg, IDCANCEL, GetLangStringW(wLangSystem, STRID_CANCEL));

    SendMessage(hWndStepEnable, BM_SETCHECK, bStepEnable, 0);
    SendMessage(hWndStepTime, EM_LIMITTEXT, 5, 0);
    SendMessage(hWndStepWidth, EM_LIMITTEXT, 5, 0);
    SetDlgItemInt(hDlg, IDC_STEPTIME, dwStepTime, FALSE);
    SetDlgItemInt(hDlg, IDC_STEPWIDTH, nStepWidth, TRUE);

    if (dwHSync & SNC_MDI)
      SendMessage(hWndHSyncMDI, BM_SETCHECK, BST_CHECKED, 0);
    if (dwHSync & SNC_CLONE)
      SendMessage(hWndHSyncClone, BM_SETCHECK, BST_CHECKED, 0);
    if (dwVSync & SNC_MDI)
      SendMessage(hWndVSyncMDI, BM_SETCHECK, BST_CHECKED, 0);
    if (dwVSync & SNC_CLONE)
      SendMessage(hWndVSyncClone, BM_SETCHECK, BST_CHECKED, 0);

    if (dwNoScroll & NS_UNDO)
      SendMessage(hWndUndo, BM_SETCHECK, BST_CHECKED, 0);
    if (dwNoScroll & NS_REDO)
      SendMessage(hWndRedo, BM_SETCHECK, BST_CHECKED, 0);
    if (dwNoScroll & NS_SELECTALL)
      SendMessage(hWndSelectAll, BM_SETCHECK, BST_CHECKED, 0);

    if (!bAkelEdit || !bMDI)
    {
      EnableWindow(hWndHSyncMDI, FALSE);
      EnableWindow(hWndVSyncMDI, FALSE);
    }
    if (!bAkelEdit)
    {
      EnableWindow(hWndHSyncClone, FALSE);
      EnableWindow(hWndVSyncClone, FALSE);

      EnableWindow(hWndUndo, FALSE);
      EnableWindow(hWndRedo, FALSE);
      EnableWindow(hWndSelectAll, FALSE);
    }
    SendMessage(hDlg, WM_COMMAND, IDC_STEPENABLE, 0);
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_STEPENABLE)
    {
      bState=SendMessage(hWndStepEnable, BM_GETCHECK, 0, 0);
      EnableWindow(hWndStepTime, bState);
      EnableWindow(hWndStepWidth, bState);
    }
    else if (LOWORD(wParam) == IDOK)
    {
      bStepEnable=SendMessage(hWndStepEnable, BM_GETCHECK, 0, 0);
      dwStepTime=GetDlgItemInt(hDlg, IDC_STEPTIME, NULL, FALSE);
      nStepWidth=GetDlgItemInt(hDlg, IDC_STEPWIDTH, NULL, TRUE);

      dwHSync=0;
      if (SendMessage(hWndHSyncMDI, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwHSync|=SNC_MDI;
      if (SendMessage(hWndHSyncClone, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwHSync|=SNC_CLONE;

      dwVSync=0;
      if (SendMessage(hWndVSyncMDI, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwVSync|=SNC_MDI;
      if (SendMessage(hWndVSyncClone, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwVSync|=SNC_CLONE;

      dwNoScroll=0;
      if (SendMessage(hWndUndo, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwNoScroll|=NS_UNDO;
      if (SendMessage(hWndRedo, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwNoScroll|=NS_REDO;
      if (SendMessage(hWndSelectAll, BM_GETCHECK, 0, 0) == BST_CHECKED)
        dwNoScroll|=NS_SELECTALL;

      SaveOptionsW();
      EndDialog(hDlg, 0);

      if (bInitMain)
      {
        if (!hWndScrolling)
        {
          EDITINFO ei;

          if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
            hWndScrolling=ei.hWndEdit;
        }
        UninitMain();
        InitMain();
      }
      return TRUE;
    }
    else if (LOWORD(wParam) == IDCANCEL)
    {
      EndDialog(hDlg, 0);
      return TRUE;
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }
  return FALSE;
}


void SetScrollTimer()
{
  if (bStepEnable)
  {
    if (hWndScrolling)
    {
      HideCaret(hWndScrolling);
      dwScrollTimer=SetTimer(NULL, 0, dwStepTime, (TIMERPROC)TimerProc);
    }
  }
}

void KillScrollTimer()
{
  if (dwScrollTimer)
  {
    KillTimer(NULL, dwScrollTimer);
    dwScrollTimer=0;
    if (hWndScrolling) ShowCaret(hWndScrolling);
  }
}

VOID CALLBACK TimerProc(HWND hwnd, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
{
  if (hWndScrolling)
  {
    SCROLLINFO si;
    POINT pt;

    si.cbSize=sizeof(SCROLLINFO);
    si.fMask=SIF_PAGE|SIF_POS|SIF_RANGE;
    GetScrollInfo(hWndScrolling, SB_VERT, &si);

    if (si.nPage > 0 && si.nPos + (int)si.nPage < si.nMax)
    {
      SendMessage(hWndScrolling, EM_GETSCROLLPOS, 0, (LPARAM)&pt);
      pt.y+=nStepWidth;
      SendMessage(hWndScrolling, EM_SETSCROLLPOS, 0, (LPARAM)&pt);
    }
  }
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="StepEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bStepEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="StepTime";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwStepTime;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="StepWidth";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nStepWidth;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="HSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwHSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="VSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwVSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NoScroll";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNoScroll;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"StepEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bStepEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"StepTime";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwStepTime;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"StepWidth";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nStepWidth;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"HSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwHSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"VSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwVSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NoScroll";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNoScroll;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    po.pOptionName="StepEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bStepEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="StepTime";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwStepTime;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="StepWidth";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nStepWidth;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="HSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwHSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="VSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwVSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="NoScroll";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNoScroll;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"StepEnable";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&bStepEnable;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"StepTime";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwStepTime;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"StepWidth";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nStepWidth;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"HSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwHSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"VSync";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwVSync;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"NoScroll";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&dwNoScroll;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_AUTOSTEP)
      return L"\x0410\x0432\x0442\x043E\x043C\x0430\x0442\x0438\x0447\x0435\x0441\x043A\x0430\x044F\x0020\x043F\x0440\x043E\x043A\x0440\x0443\x0442\x043A\x0430";
    if (nStringID == STRID_ENABLE)
      return L"\x0412\x043A\x043B\x044E\x0447\x0435\x043D\x043E";
    if (nStringID == STRID_STEPTIME)
      return L"\x0412\x0440\x0435\x043C\x044F\x0020\x0448\x0430\x0433\x0430";
    if (nStringID == STRID_STEPWIDTH)
      return L"\x0428\x0438\x0440\x0438\x043D\x0430\x0020\x0448\x0430\x0433\x0430";
    if (nStringID == STRID_HSYNCHRONIZATION)
      return L"\x0413\x043E\x0440\x0438\x0437\x043E\x043D\x0442\x0430\x043B\x044C\x043D\x0430\x044F\x0020\x0441\x0438\x043D\x0445\x0440\x043E\x043D\x0438\x0437\x0430\x0446\x0438\x044F";
    if (nStringID == STRID_VSYNCHRONIZATION)
      return L"\x0412\x0435\x0440\x0442\x0438\x043A\x0430\x043B\x044C\x043D\x0430\x044F\x0020\x0441\x0438\x043D\x0445\x0440\x043E\x043D\x0438\x0437\x0430\x0446\x0438\x044F";
    if (nStringID == STRID_FRAMEMDI)
      return L"\x0412\x043A\x043B\x0430\x0434\x043A\x0438\x0020\x0028\x004D\x0044\x0049\x0029";
    if (nStringID == STRID_SPLITPANE)
      return L"\x0420\x0430\x0437\x0434\x0435\x043B\x0435\x043D\x043D\x044B\x0435\x0020\x043E\x043A\x043D\x0430";
    if (nStringID == STRID_NOSCROLL)
      return L"\x041D\x0435\x0020\x043F\x0440\x043E\x043A\x0440\x0443\x0447\x0438\x0432\x0430\x0442\x044C\x0020\x043E\x043F\x0435\x0440\x0430\x0446\x0438\x0438";
    if (nStringID == STRID_UNDO)
      return L"\x041E\x0442\x043C\x0435\x043D\x0438\x0442\x044C\x0020\x0028\x0432\x0435\x0441\x044C\x0020\x0442\x0435\x043A\x0441\x0442\x0029";
    if (nStringID == STRID_REDO)
      return L"\x041F\x043E\x0432\x0442\x043E\x0440\x0438\x0442\x044C\x0020\x0028\x0432\x0435\x0441\x044C\x0020\x0442\x0435\x043A\x0441\x0442\x0029";
    if (nStringID == STRID_SELECTALL)
      return L"\x0412\x044B\x0434\x0435\x043B\x0438\x0442\x044C\x0020\x0432\x0441\x0435";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
    if (nStringID == STRID_CANCEL)
      return L"\x041E\x0442\x043C\x0435\x043D\x0430";
  }
  else
  {
    if (nStringID == STRID_AUTOSTEP)
      return L"Automatic scrolling";
    if (nStringID == STRID_ENABLE)
      return L"Enable";
    if (nStringID == STRID_STEPTIME)
      return L"Step time";
    if (nStringID == STRID_STEPWIDTH)
      return L"Step width";
    if (nStringID == STRID_HSYNCHRONIZATION)
      return L"Horizontal synchronization";
    if (nStringID == STRID_VSYNCHRONIZATION)
      return L"Vertical synchronization";
    if (nStringID == STRID_FRAMEMDI)
      return L"Tabs (MDI)";
    if (nStringID == STRID_SPLITPANE)
      return L"Split panes";
    if (nStringID == STRID_NOSCROLL)
      return L"No scroll operations";
    if (nStringID == STRID_UNDO)
      return L"Undo (all text)";
    if (nStringID == STRID_REDO)
      return L"Redo (all text)";
    if (nStringID == STRID_SELECTALL)
      return L"Select all";
    if (nStringID == STRID_OK)
      return L"OK";
    if (nStringID == STRID_CANCEL)
      return L"Cancel";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  hMdiClient=pd->hMdiClient;
  hMainIcon=pd->hMainIcon;
  bOldWindows=pd->bOldWindows;
  bMDI=pd->bMDI;
  bAkelEdit=pd->bAkelEdit;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
    ReadOptionsA();
  else
    ReadOptionsW();
}

void InitMain()
{
  bInitMain=TRUE;

  //Subclass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);

  if (bMDI)
  {
    NewFrameProcData=NULL;
    SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NewFrameProc, (LPARAM)&NewFrameProcData);
  }

  //Set timer
  SetScrollTimer();
}

void UninitMain()
{
  bInitMain=FALSE;

  //Remove timer
  KillScrollTimer();

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }
  if (bMDI)
  {
    if (NewFrameProcData)
    {
      SendMessage(hMainWnd, AKD_SETFRAMEPROC, (WPARAM)NULL, (LPARAM)&NewFrameProcData);
      NewFrameProcData=NULL;
    }
  }
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
